import os
import re

from setuptools import setup, find_packages

readme = os.path.join(os.path.dirname(__file__), 'README.rst')

setup(
    name="pynuoca", # Replace with your own username
    version="20.0.1",
    author="NuoDB",
    author_email="support@nuodb.com",
    description="NuoDB Collection Agent (NuoCA)",
    long_description=open(readme).read(),
    url="https://github.com/nuodb/nuoca",
    license='MIT',
    keywords='nuodb scalable cloud database',
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests", "tests/*"]),
    include_package_data=True,
    install_requires=[
        'aenum>=2.0.8',
        'click>=6.7',
        'elasticsearch>=5.0.0,<6.0.0',
        'python-dateutil>=2.6.1',
        'PyPubSub>=3.3.0',
        'PyYaml>=3.12',
        'requests>=2.21.0',
        'wrapt>=1.10.11',
        'Yapsy>=1.11.223',
        'kafka-python>=1.4.1',
        'pynuoadmin>=1.0.0',
        'pynuodb>=2.3.2',
    ],
    scripts=['bin/nuoca',
    ],
    data_files=[
        ('etc', ['etc/nuoca_setup.sh', 'etc/nuoca_export.sh', 'etc/nuoca_env.sh',
                 'etc/nuoca_settings.yml', 'etc/nuoca.yml']),
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX",
        'Topic :: Database :: Front-Ends',
    ],
)



# (C) Copyright NuoDB, Inc. 2017-2020
#
# This source code is licensed under the MIT license found in the LICENSE
# file in the root directory of this source tree.

import logging
import os
import re
import socket
import threading
import time
import datetime

from copy import deepcopy
from pynuoca.nuoca_plugin import NuocaMPInputPlugin
from pynuoca.nuoca_util import nuoca_log

try:
    from pynuoadmin import nuodb_mgmt
except ImportError:
    import nuodb_mgmt

# NuoAdminNuoMonitor plugin
#
# NOTE: This is an Alpha version of a NuoCA plugin that can collect NuoDB
# Engine metrics (NuoMon) from the NuoDB new Admin layer (which is a Beta
# feature for NuoDB release 3.2). This NuoCA plugin has minimal testing and
# is expected to change (or be incorporated into the NuoCA NuoMon plugin)
#
# Example NuoAdminNuoMonitor plugin configuration:
#
# - NuoAdminNuoMonitor:
#    description : Collection from internal nuoAdminNuoMonitor tool
#    database_regex_pattern: dbt2
#    api_server: localhost:8888
#    client_key: /opt/nuodb/tls-config/keys/nuocmd.pem
#    server_cert: None
#    server_id: nuoadmin0

class BaseCollector(object):

    def message_received(self, root):
        pass

    def invalid_message(self, message):
        pass

    def closed(self):
        pass


class BaseMetricsCollector(BaseCollector):
    def __init__(self):
        super(BaseMetricsCollector, self).__init__()
        self.__first = True
        self.__process = None

    @property
    def process(self):
        return self.__process

    @process.setter
    def process(self, p):
        self.__process = p

    def __get_item(self, attrs):
        units = ["COUNT", "MILLISECONDS", "STATE",
                 "NUMBER", "PERCENT", "IDENTIFIER",
                 "DELTA"]
        return {"unit": units[int(attrs['units']) - 1],
                "description": attrs['header']}

    # @trace
    def message_received(self, root):
        def parseStr(x):
            try:
                return int(x)
            except:
                return x

        items = {}
        if root.tag == "Items":
            for child in root:
                items[child.attrib['name']] = self.__get_item(child.attrib)
            items['Database'] = dict(unit="IDENTIFIER", description="Database Name")
            items['Region'] = dict(unit="IDENTIFIER", description="Region Name")
            self.onStart(items)
            return None
        elif root.tag == 'Status':
            new_values = dict([(k, parseStr(v)) for k, v in root.attrib.iteritems()])
            if self.__first:
                new_values['Database'] = self.process.db_name
                new_values['Region'] = self.process.region_name
                self.__first = False
            new_values['TimeStamp'] = int(time.time() * 1000.0)
            self.onChange(new_values)
            return None

    def closed(self):
        self.onEnd()
        pass


class MetricsCollector(BaseMetricsCollector):
    """ Base class for metrics collection.
    Remembers previous values"""

    def __init__(self):
        super(MetricsCollector, self).__init__()
        self.__metrics = {}
        self.__values = {}
        pass

    @property
    def metrics(self):
        return self.__metrics

    @property
    def values(self):
        return self.__values

    def init(self, args):
        return self

    def onStart(self, metrics):
        """ remembers metrics  """
        self.__metrics = metrics

    def onChange(self, metrics):
        """ remembers previous values """
        self.__values.update(metrics)

    def onEnd(self):
        """ zero all values """
        zeroMetrics = {}
        for k, v in self.__values.iteritems():
            if v != 0 and type(v) is int:
                zeroMetrics[k] = 0
        zeroMetrics['TimeStamp'] = int(time.time() * 1000.0)
        self.onChange(zeroMetrics)
        pass


class NuoAdminNuoMonMessageConsumer(object):
    """ NuoAdminNuoMonMessageConsumer"""

    _conn = None
    _nuo_monitor_obj = None

    def __init__(self, conn, nuo_monitor_obj):
        self._conn = conn
        self._nuo_monitor_obj = nuo_monitor_obj
        self._process_metrics_dict = {}

    def get_stats(self, db_name=None, start_id=None, server_id=None):
        while self._nuo_monitor_obj._enabled:
            try:
                for process_msg in self._get_messages(None, db_name,
                                                      start_id, server_id):
                    if process_msg and 'msg' in process_msg:
                        self._nuo_monitor_obj.nuoAdminNuoMonitor_collect_queue.append(
                            deepcopy(process_msg['msg']))
            except Exception as e:
                nuoca_log(logging.ERROR, "Exception in NuoAdminNuoMonMessage"
                                         "ConsumerNuoAdminNuoMon.get_stats: %s"
                          % str(e))
            time.sleep(10)

    def _get_messages(self, log_options, db_name, start_id, server_id,
                      include_process=False):

        # message generator is for a single process if `start_id` is specified;
        # otherwise, it is aggregated
        if start_id is not None:
            msg_stream = self._conn.monitor_process(start_id, log_options)
        else:
            msg_stream = self._conn.monitor_database(db_name, log_options,
                                                     keep_open=True)
        # filter messages by name based on whether we are streaming stats or
        # log messages
        message_name = 'Status' if log_options is None else 'LogMessage'
        for process, xml_message in msg_stream:
            if server_id:
                if process.server_id != server_id:
                    continue
            if process.start_id not in self._process_metrics_dict:
                mc = MetricsCollector()
                self._process_metrics_dict[process.start_id] = mc
                mc.process = process
            else:
                mc = self._process_metrics_dict[process.start_id]
            mc.message_received(xml_message)
            items = mc.values
            json_msg = items
            if len(json_msg) != 0:
                # add timestamp to message attributes; TODO: in the future, we
                # may want to include timestamp on the sending side
                if 'Time' not in json_msg:
                    json_msg['Time'] = self._get_timestamp()
                # combine process and message; if `include_process`, return all
                # process attributes; otherwise just return start ID

                if include_process:
                    process._dict['msg'] = json_msg
                    yield process
                else:
                    yield dict(startId=process.start_id, msg=json_msg)

    @staticmethod
    def _get_timestamp(time_sec=None):
        if time_sec is None:
            time_sec = time.time()
        dt = datetime.datetime.fromtimestamp(time_sec)
        return dt.strftime('%Y-%m-%dT%H:%M:%S.%f')


class NuoAdminNuoMonitorPlugin(NuocaMPInputPlugin):
    DEFAULT_API_SERVER = 'localhost:8888'

    def __init__(self, parent_pipe):
        super(NuoAdminNuoMonitorPlugin, self).__init__(parent_pipe, 'NuoAdminNuoMon')
        self._config = None
        self._nuocaCollectionName = None
        self._api_server = NuoAdminNuoMonitorPlugin.DEFAULT_API_SERVER
        self._server_id = None
        self._client_key = None
        self._server_cert = None
        self._enabled = False
        self._numon_handler_ready = False
        self._domain = None
        self._domain_metrics = None
        self._domain_metrics_host = None
        self._database_regex_pattern = '.*'
        self._host_uuid_shortname = False
        self._thread = None
        self._nuoAdminNuoMonitor_collect_queue = []
        self._collection_interval = 30

    def _get_admin_conn(self):
        client_key = self._client_key
        if client_key is not None and ',' in client_key:
            client_key = client_key.split(',')
            if len(client_key) != 2:
                raise ValueError(
                    'Expected at most two tokens for --client-key')
        server_cert = self._server_cert
        api_server = self._api_server
        if (not api_server.startswith('http://') and
                not api_server.startswith('https://')):  # noqa
            if client_key is None and server_cert is None:
                api_server = 'http://' + api_server
            else:
                api_server = 'https://' + api_server
        verify = server_cert
        if not verify:
            verify = False

        return nuodb_mgmt.AdminConnection(api_server, client_key, verify)

    def wait_for_terminate(self):
        if self._domain:
            self._domain.waitForTerminate()

    @property
    def nuoAdminNuoMonitor_collect_queue(self):
        return self._nuoAdminNuoMonitor_collect_queue

    def _NuoAdminNuoMon_handler_thread(self):
        self._numon_handler_ready = True
        self._domain_metrics.get_stats(server_id=self._server_id)

    def startup_NuoAdminNuoMon(self):
        try:
            self._numon_handler_ready = False
            self._conn = self._get_admin_conn()
            self._enabled = True
            self._domain_metrics = NuoAdminNuoMonMessageConsumer(self._conn, self)
            self._thread = threading.Thread(target=self._NuoAdminNuoMon_handler_thread)
            self._thread.daemon = True
            self._thread.start()
            try_count = 0
            while not self._numon_handler_ready and try_count < 5:
                try_count += 1
                time.sleep(1)
            return self._numon_handler_ready
        except Exception as e:
            nuoca_log(logging.ERROR, "NuoAdminNuoMon Startup error: %s" % str(e))
            return False

    def startup(self, config=None):
        try:
            self._config = config

            # Validate the configuration.
            required_config_items = ['api_server']
            if not self.has_required_config_items(config, required_config_items):
                return False

            # Don't reveal the domain password in the NuoCA log file.
            display_config = {}
            display_config.update(config)
            display_config['domain_password'] = ''
            nuoca_log(logging.INFO, "NuoAdminNuoMon: plugin config: %s" %
                      str(display_config))

            self._api_server = os.path.expandvars(config['api_server'])
            if 'server_id' in config:
                self._server_id = os.path.expandvars(config['server_id'])
            if 'client_key' in config:
                self._client_key = os.path.expandvars(config['client_key'])
            if 'server_cert' in config:
                self._server_cert = os.path.expandvars(config['server_cert'])
            if 'nuocaCollectionName' in config:
                self._nuocaCollectionName = config['nuocaCollectionName']
            if 'domain_metrics_host' in config:
                self._domain_metrics_host = os.path.expandvars(config['domain_metrics_host'])
                if self._domain_metrics_host == 'localhost':
                    self._domain_metrics_host = socket.gethostname()
            if 'database_regex_pattern' in config:
                self._database_regex_pattern = config['database_regex_pattern']
            if 'host_uuid_shortname' in config:
                self._host_uuid_shortname = config['host_uuid_shortname']
            self.startup_NuoAdminNuoMon()
            return self._numon_handler_ready
        except Exception as e:
            nuoca_log(logging.ERROR, "NuoAdminNuoMon Plugin Startup error: %s" % str(e))
            return False

    def shutdown(self):
        self._enabled = False
        pass

    def collect(self, collection_interval):
        uuid_hostname_regex = \
            '^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}-'
        rval = None
        self._collection_interval = collection_interval
        try:
            nuoca_log(logging.DEBUG, "NuoAdminNuoMon: collect()")
            base_values = super(NuoAdminNuoMonitorPlugin, self).collect(collection_interval)
            collection_count = len(self._nuoAdminNuoMonitor_collect_queue)
            nuoca_log(logging.DEBUG, "NuoAdminNuoMon: collection_count %s" %
                      str(collection_count))
            if not collection_count:
                return rval

            rval = []
            for i in range(collection_count):
                collected_dict = self._nuoAdminNuoMonitor_collect_queue.pop(0)
                if self._domain_metrics_host and 'Hostname' in collected_dict:
                    if collected_dict['Hostname'] != self._domain_metrics_host:
                        continue
                if self._nuocaCollectionName:
                    collected_dict['nuocaCollectionName'] = self._nuocaCollectionName
                if 'Database' in collected_dict:
                    m = re.search(self._database_regex_pattern, collected_dict['Database'])
                    if m:
                        if self._host_uuid_shortname:
                            m2 = re.search(uuid_hostname_regex, collected_dict['Hostname'])
                            if m2:
                                shortid = collected_dict['Hostname'][37:]
                                if 'NodeType' in collected_dict:
                                    if collected_dict['NodeType'] == 'Transaction':
                                        shortid += "(TE)"
                                    elif collected_dict['NodeType'] == 'Storage':
                                        shortid += "(SM)"
                                shortid_with_pid = shortid + str(collected_dict['ProcessId'])
                                collected_dict['HostShortID'] = shortid
                                collected_dict['HostShortIDwithPID'] = shortid_with_pid
                        rval.append(collected_dict)
                else:
                    rval.append(collected_dict)

        except Exception as e:
            nuoca_log(logging.ERROR, str(e))
        return rval



# (C) Copyright NuoDB, Inc. 2017-2020
#
# This source code is licensed under the MIT license found in the LICENSE
# file in the root directory of this source tree.

import os
import logging
import json

from pynuoca.nuoca_plugin import NuocaMPOutputPlugin
from pynuoca.nuoca_util import nuoca_log


class FilePlugin(NuocaMPOutputPlugin):
    def __init__(self, parent_pipe, config=None):
        super(FilePlugin, self).__init__(parent_pipe, 'File')
        self._config = config
        self._file_path = None
        self._fp = None

    def startup(self, config=None):
        try:
            self._config = config
            required_config_items = ['filePath']
            if not self.has_required_config_items(config, required_config_items):
                return False
            self._file_path = os.path.expandvars(config['filePath'])
            self._fp = open(self._file_path, 'w')
            return True
        except Exception as e:
            nuoca_log(logging.ERROR, str(e))
            return False

    def shutdown(self):
        if self._fp:
            self._fp.close()
            self._fp = None

    def store(self, ts_values):
        rval = None
        try:
            nuoca_log(logging.DEBUG, "Called store() in MPCounterPlugin process")
            rval = super(FilePlugin, self).store(ts_values)
            json.dump(ts_values, self._fp)
            self._fp.write("\n")

        except Exception as e:
            nuoca_log(logging.ERROR, str(e))
        return rval



# (C) Copyright NuoDB, Inc. 2017-2020
#
# This source code is licensed under the MIT license found in the LICENSE
# file in the root directory of this source tree.

from contextlib import closing

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

metrics = {}


def MONITOR_COUNT(k, values):
    raw = values[k]
    return "raw=%si" % (raw)


def MONITOR_DELTA(k, values):
    raw = int(values[k])
    rvalues = "raw=%di" % (raw)
    if 'Milliseconds' in values:
        ms = int(values['Milliseconds'])
        rate = raw * 1000. / ms if ms != 0 else 0.
        rvalues += ",rate=%f" % (rate)
    return rvalues


def MONITOR_MILLISECONDS(k, values):
    raw = int(values[k])
    rvalues = "raw=%di" % (raw)
    if 'Milliseconds' in values:
        ms = int(values['Milliseconds'])
        value = raw * 1. / ms if ms != 0 else 0.
        rvalues += ",value=%f" % (value)
        if 'NumberCores' in values:
            ncores = int(values['NumberCores'])
            nvalue = value / ncores if ncores != 0 else 0.
            rvalues += ",normvalue=%f" % (nvalue)

    return rvalues


def MONITOR_NUMBER(k, values):
    raw = values[k]
    return "raw=%si" % (raw)


MONITOR_IDENTIFIER = None


def MONITOR_PERCENT(k, values):
    raw = values[k]
    rvalues = "raw=%si" % (raw)
    if 'NumberCores' in values:
        ncores = int(values['NumberCores'])
        if ncores != 0:
            value = int(raw) * 1. / ncores
            bycore = int(raw) * .01
            # norm   -> 0 <-> 100 * ncores
            # ncores -> 0 <-> ncores
            rvalues += ",norm=%f,ncores=%f" % (value, bycore)
            if k == 'PercentCpuTime':
                idle = 100 * ncores - int(raw)
                # idle   -> 0 <-> 100 * ncores
                # nidle  -> 0 <-> ncores
                rvalues += ",idle=%di,nidle=%f" % (idle, float(idle) / ncores)
        else:
            # ncores only 0 at disconnect
            rvalues += ",norm=%f,ncores=%f" % (0., 0.)
            if k == 'PercentCpuTime':
                rvalues += ",idle=%di,nidle=%f" % (0., 0.)
    return rvalues


def summary(values):
    # from nuodbmgr
    def get(key):
        return int(values[key]) if key in values else 0

    activeTime = get("ActiveTime")
    deltaTime = get("Milliseconds")
    idleTime = get("IdleTime")

    def v(raw):
        rvalues = "raw=%d" % (raw)
        if deltaTime > 0:
            rvalues += ",nthreads=%d" % (int(round(raw / deltaTime)))
            if activeTime > 0:
                multiplier = (deltaTime - idleTime) * 1. / deltaTime
                rvalues += ",percent=%d" % (int(round(multiplier * raw * 100. / activeTime)))
        return rvalues

    cpuTime = get("UserMilliseconds") + get("KernelMilliseconds")
    syncTime = get("SyncPointWaitTime") + get("StallPointWaitTime")
    syncTime -= get("PlatformObjectCheckOpenTime") + get("PlatformObjectCheckPopulatedTime") + get(
        "PlatformObjectCheckCompleteTime")
    lockTime = get("TransactionBlockedTime")
    fetchTime = get("PlatformObjectCheckOpenTime") + get("PlatformObjectCheckPopulatedTime") + get(
        "PlatformObjectCheckCompleteTime") + get("LoadObjectTime")
    commitTime = get("RemoteCommitTime")
    ntwkSendTime = get("NodeSocketBufferWriteTime")
    archiveReadTime = get("ArchiveReadTime")
    archiveWriteTime = get("ArchiveWriteTime") + get("ArchiveFsyncTime") + get("ArchiveDirectoryTime")
    journalWriteTime = get("JournalWriteTime") + get("JournalFsyncTime") + get("JournalDirectoryTime")
    throttleTime = get("ArchiveSyncThrottleTime") + get("MemoryThrottleTime") + get("WriteThrottleTime")
    throttleTime += get("ArchiveBandwidthThrottleTime") + get("JournalBandwidthThrottleTime")
    values = {
        "Summary.Active": v(activeTime),
        "Summary.CPU": v(cpuTime),
        "Summary.Idle": v(idleTime),
        "Summary.Sync": v(syncTime),
        "Summary.Lock": v(lockTime),
        "Summary.Fetch": v(fetchTime),
        "Summary.Commit": v(commitTime),
        "Summary.NtwkSend": v(ntwkSendTime),
        "Summary.ArchiveRead": v(archiveReadTime),
        "Summary.ArchiveWrite": v(archiveWriteTime),
        "Summary.JournalWrite": v(journalWriteTime),
        "Summary.Throttle": v(throttleTime)
    }
    return values


metrics = {
    "Objects": MONITOR_COUNT,
    "PurgedObjects": MONITOR_COUNT,
    "ObjectsCreated": MONITOR_DELTA,
    "ObjectsSaved": MONITOR_DELTA,
    "ObjectsImported": MONITOR_DELTA,
    "ObjectsExported": MONITOR_DELTA,
    "ObjectsLoaded": MONITOR_DELTA,
    "ObjectsReloaded": MONITOR_DELTA,
    "ObjectsPurged": MONITOR_DELTA,
    "ObjectsDropped": MONITOR_DELTA,
    "ObjectsDroppedPurged": MONITOR_DELTA,
    "ObjectsDeleted": MONITOR_DELTA,
    "ObjectsRequested": MONITOR_DELTA,
    "ObjectsBounced": MONITOR_DELTA,
    "Commits": MONITOR_DELTA,
    "Rollbacks": MONITOR_DELTA,
    "Inserts": MONITOR_DELTA,
    "Updates": MONITOR_DELTA,
    "Deletes": MONITOR_DELTA,
    "Stalls": MONITOR_DELTA,
    "SqlMsgs": MONITOR_DELTA,
    "SqlListenerSqlProcTime": MONITOR_MILLISECONDS,
    "SqlListenerThrottleTime": MONITOR_MILLISECONDS,
    "SqlListenerIdleTime": MONITOR_MILLISECONDS,
    "SqlListenerStallTime": MONITOR_MILLISECONDS,
    "NodePingTime": MONITOR_MILLISECONDS,
    "NodeApplyPingTime": MONITOR_MILLISECONDS,
    "NodePostMethodTime": MONITOR_MILLISECONDS,
    "IdManagerBlockingStallCount": MONITOR_DELTA,
    "IdManagerNonBlockingStallCount": MONITOR_DELTA,
    "PendingUpdateStallCount": MONITOR_DELTA,
    "PlatformCatalogStallCount": MONITOR_DELTA,
    "PlatformObjectCheckOpenTime": MONITOR_MILLISECONDS,
    "PlatformObjectCheckPopulatedTime": MONITOR_MILLISECONDS,
    "PlatformObjectCheckCompleteTime": MONITOR_MILLISECONDS,
    "PlatformIndexCheckAcknowledgedTime": MONITOR_MILLISECONDS,
    "StallPointWaitTime": MONITOR_MILLISECONDS,
    "SyncPointWaitTime": MONITOR_MILLISECONDS,
    "TransactionBlockedTime": MONITOR_MILLISECONDS,
    "PendingInsertWaitTime": MONITOR_MILLISECONDS,
    "PendingUpdateWaitTime": MONITOR_MILLISECONDS,
    "MessagesReceived": MONITOR_DELTA,
    "MessagesSent": MONITOR_DELTA,
    "PacketsReceived": MONITOR_DELTA,
    "PacketsSent": MONITOR_DELTA,
    "AdminReceived": MONITOR_DELTA,
    "AdminSent": MONITOR_DELTA,
    "ClientReceived": MONITOR_DELTA,
    "ClientSent": MONITOR_DELTA,
    "ServerReceived": MONITOR_DELTA,
    "ServerSent": MONITOR_DELTA,
    "BytesReceived": MONITOR_DELTA,
    "BytesSent": MONITOR_DELTA,
    "BytesBuffered": MONITOR_DELTA,
    "DiskWritten": MONITOR_DELTA,
    "HeapAllocated": MONITOR_NUMBER,
    "HeapActive": MONITOR_NUMBER,
    "HeapMapped": MONITOR_NUMBER,
    "Memory": MONITOR_NUMBER,
    "PageFaults": MONITOR_COUNT,
    "ArchiveQueue": MONITOR_COUNT,
    "LogMsgs": MONITOR_COUNT,
    "JrnlWrites": MONITOR_DELTA,
    "JrnlBytes": MONITOR_DELTA,
    "JournalQueue": MONITOR_COUNT,
    "RefactorTXQueueTime": MONITOR_MILLISECONDS,
    "NodeSocketBufferWriteTime": MONITOR_MILLISECONDS,
    "MessageSequencerSortTime": MONITOR_MILLISECONDS,
    "MessageSequencerMergeTime": MONITOR_MILLISECONDS,
    "BroadcastTime": MONITOR_MILLISECONDS,
    "NonChairSplitTime": MONITOR_MILLISECONDS,
    "WaitForSplitTime": MONITOR_MILLISECONDS,
    "NumSplits": MONITOR_DELTA,
    "ArchiveBufferedBytes": MONITOR_DELTA,
    "Milliseconds": MONITOR_MILLISECONDS,
    "UserMilliseconds": MONITOR_MILLISECONDS,
    "KernelMilliseconds": MONITOR_MILLISECONDS,
    "NumberCores": MONITOR_NUMBER,
    "Hostname": MONITOR_IDENTIFIER,
    "NodeType": MONITOR_IDENTIFIER,
    "NodeState": MONITOR_IDENTIFIER,
    "ArchiveDirectory": MONITOR_IDENTIFIER,
    "PercentUserTime": MONITOR_PERCENT,
    "PercentSystemTime": MONITOR_PERCENT,
    "PercentCpuTime": MONITOR_PERCENT,
    "EffectiveVersion": MONITOR_NUMBER,
    "ActualVersion": MONITOR_NUMBER,
    "ProcessId": MONITOR_IDENTIFIER,
    "NodeId": MONITOR_IDENTIFIER,
    "ClientCncts": MONITOR_NUMBER,
    "CurrentCommittedTransactions": MONITOR_NUMBER,
    "CurrentActiveTransactions": MONITOR_NUMBER,
    "CurrentPurgedTransactions": MONITOR_NUMBER,
    "OldestActiveTransaction": MONITOR_NUMBER,
    "OldestCommittedTransaction": MONITOR_NUMBER,
    "SnapshotAlbumsClosed": MONITOR_DELTA,
    "SnapshotAlbumsClosedForGC": MONITOR_DELTA,
    "SnapshotAlbumTime": MONITOR_NUMBER,
    "SnapshotAlbumSize": MONITOR_NUMBER,
    "SendQueueSize": MONITOR_NUMBER,
    "SendSortingQueueSize": MONITOR_NUMBER,
    "SocketBufferBytes": MONITOR_NUMBER,
    "RemoteCommitTime": MONITOR_MILLISECONDS,
    "IdleTime": MONITOR_MILLISECONDS,
    "ActiveTime": MONITOR_MILLISECONDS,
    "ClientThreadBacklog": MONITOR_MILLISECONDS,
    "AtomProcessingThreadBacklog": MONITOR_MILLISECONDS,
    "HTTPProcessingThreadBacklog": MONITOR_MILLISECONDS,
    "ArchiveReadTime": MONITOR_MILLISECONDS,
    "ArchiveWriteTime": MONITOR_MILLISECONDS,
    "ArchiveFsyncTime": MONITOR_MILLISECONDS,
    "ArchiveDirectoryTime": MONITOR_MILLISECONDS,
    "JournalWriteTime": MONITOR_MILLISECONDS,
    "JournalFsyncTime": MONITOR_MILLISECONDS,
    "JournalDirectoryTime": MONITOR_MILLISECONDS,
    "LoadObjectTime": MONITOR_MILLISECONDS,
    "ArchiveSyncThrottleTime": MONITOR_MILLISECONDS,
    "MemoryThrottleTime": MONITOR_MILLISECONDS,
    "PruneAtomsThrottleTime": MONITOR_MILLISECONDS,
    "ArchiveBandwidthThrottleTime": MONITOR_MILLISECONDS,
    "JournalBandwidthThrottleTime": MONITOR_MILLISECONDS,
    "CreatePlatformRecordsTime": MONITOR_MILLISECONDS,
    "PendingEventsCommitTime": MONITOR_MILLISECONDS,
    "DependentCommitWaits": MONITOR_NUMBER,
    "ObjectFootprint": MONITOR_NUMBER,
    "WriteLoadLevel": MONITOR_NUMBER,
    "WriteThrottleSetting": MONITOR_NUMBER,
    "WriteThrottleTime": MONITOR_MILLISECONDS,
    "LocalCommitOrderTime": MONITOR_MILLISECONDS,
    "CheckCompleteOptimized": MONITOR_DELTA,
    "CheckCompleteFull": MONITOR_DELTA,
    "ChairmanMigration": MONITOR_DELTA,
    "CycleTime": MONITOR_MILLISECONDS
}


def format(values):
    # metric,<identity tags> <fields> timestamp
    if 'TimeStamp' not in values:
        return

    timestamp = values['TimeStamp']
    # timestamp is in milliseconds -> nanoseconds
    timestamp = int(timestamp * 1000000)
    nodetype = values['NodeType']
    hostname = values['Hostname']
    processid = values['ProcessId']
    nodeid = values['NodeId']
    database = values['Database']
    region = "<unknown>"
    if 'Region' in values:
        region = values['Region']

    header = ["TimeStamp", "NodeType", "Hostname", "ProcessId", "NodeId", "Database", "Region"]
    tags = "host=%s,nodetype=%s,pid=%s,nodeid=%s,db=%s,region=%s" % (
    hostname, nodetype, processid, nodeid, database, region)

    with closing(StringIO()) as buffer:
        for k in values:
            value_formatter = metrics[k] if k in metrics else None
            if value_formatter:
                rvalues = metrics[k](k, values)
                print >> buffer, "%s,%s %s %s" % (k, tags, rvalues, timestamp)
            elif k not in metrics and k not in header:
                # catch all if new metric added
                try:
                    rvalues = "raw=%di" % (int(values[k]))
                    print >> buffer, "%s,%s %s %s" % (k, tags, rvalues, timestamp)
                except:
                    pass
        summary_map = summary(values)
        for key, rvalues in summary_map.iteritems():
            print >> buffer, "%s,%s %s %s" % (key, tags, rvalues, timestamp)
        toinflux = buffer.getvalue()

    return ("text/plain; charset=us-ascii", toinflux)
